# /// script
# dependencies = ["tomlkit"]
# ///

import re
import subprocess
from pathlib import Path

import tomlkit

LINT_BLACKLIST = {
    "absolute_paths",  # TODO: might be nice
    "alloc_instead_of_core",
    "allow_attributes_without_reason",
    "arbitrary_source_item_ordering",
    "arithmetic_side_effects",  # TODO: might be nice
    "as_conversions",
    "big_endian_bytes",
    "cargo_common_metadata",  # TODO: might be nice
    "cognitive_complexity",
    "doc_markdown",  # Too many false positives
    "default_numeric_fallback",
    "doc_paragraphs_missing_punctuation", # TODO: very verbose
    "else_if_without_else",
    "exhaustive_enums",
    "exhaustive_structs",
    "float_arithmetic",
    "float_cmp",
    "float_cmp_const",
    "impl_trait_in_params",
    "implicit_return",
    "indexing_slicing",
    "integer_division",
    "integer_division_remainder_used",
    "iter_over_hash_type",
    "little_endian_bytes",
    "map_err_ignore",
    "min_ident_chars",
    "missing_docs_in_private_items",
    "missing_errors_doc",
    "missing_inline_in_public_items",
    "missing_panics_doc",
    "missing_trait_methods",
    "mixed_read_write_in_expression",
    "mod_module_files",
    "module_name_repetitions",
    "modulo_arithmetic",
    "multiple_crate_versions",
    "multiple_unsafe_ops_per_block",
    "must_use_candidate",  # TODO: might be nice
    "option_option",
    "pattern_type_mismatch",
    "pub_use",
    "pub_with_shorthand",
    "question_mark_used",
    "redundant_test_prefix", # TODO: might be nice
    "ref_option", # TODO: might be nice
    "self_named_module_files",  # TODO: might be nice
    "semicolon_if_nothing_returned",  # TODO: might be nice
    "semicolon_outside_block",
    "similar_names",
    "single_call_fn",
    "single_char_lifetime_names",
    "std_instead_of_alloc",
    "std_instead_of_core",
    "shadow_reuse",
    "shadow_unrelated",
    "string_slice",  # TODO: might be nice
    "too_many_lines",
    "separated_literal_suffix",
    "unreachable",  # TODO: might be nice
    "unused_trait_names", # TODO: might be nice
    "unwrap_used",  # TODO: might be nice to use expect instead
    "wildcard_enum_match_arm",  # TODO: might be nice
    "wildcard_imports",  # TODO: might be nice
}

lints = set()
clippy_help = subprocess.check_output(["cargo", "clippy", "--", "-W", "help"], cwd=Path(__file__).parent.parent / "lib" / "oxrdf").decode()
nursery_lints = [match.group(1) for match in re.finditer('clippy::([a-z-]+)', re.search('\n *clippy::nursery *(clippy::[a-z-]+,? *)*', clippy_help).group(0))]
for match in re.finditer(r'\n *clippy::([a-z-]+)  allow   ', clippy_help):
    if match.group(1) not in nursery_lints:
        lints.add(match.group(1).replace('-', '_'))

for flag in LINT_BLACKLIST:
    if flag in lints:
        lints.remove(flag)
    else:
        print(f"Unused blacklisted flag: {flag}")

cargo_path = Path(__file__).parent.parent / "Cargo.toml"
cargo_toml = tomlkit.parse(cargo_path.read_text())
cargo_toml["workspace"]["lints"]["clippy"] = {lint: "warn" for lint in sorted(lints)}
cargo_path.write_text(tomlkit.dumps(cargo_toml))



import argparse
import ast
import importlib
import inspect
import logging
import re
import subprocess
import sys
from functools import reduce
from typing import Any, Dict, List, Mapping, Optional, Set, Tuple, Union

PROTOCOLS = """
@typing.type_check_only
class AggregateFunctionAccumulator(typing.Protocol):
    def accumulate(self, element: NamedNode | BlankNode | Literal | Triple) -> None: ...
    def finish(self) -> NamedNode | BlankNode | Literal | Triple | None: ...
"""


def path_to_type(*elements: str) -> ast.expr:
    base: ast.expr = ast.Name(id=elements[0], ctx=ast.Load())
    for e in elements[1:]:
        base = ast.Attribute(value=base, attr=e, ctx=ast.Load())
    return base


OBJECT_MEMBERS = dict(inspect.getmembers(object))
BUILTINS: Dict[str, Union[None, Tuple[List[ast.expr], ast.expr]]] = {
    "__annotations__": None,
    "__bool__": ([], path_to_type("bool")),
    "__bytes__": ([], path_to_type("bytes")),
    "__class__": None,
    "__contains__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__del__": None,
    "__delattr__": ([path_to_type("str")], path_to_type("None")),
    "__delitem__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__dict__": None,
    "__dir__": None,
    "__doc__": None,
    "__eq__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__format__": ([path_to_type("str")], path_to_type("str")),
    "__ge__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__getattribute__": ([path_to_type("str")], path_to_type("typing", "Any")),
    "__getitem__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__gt__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__hash__": ([], path_to_type("int")),
    "__init__": ([], path_to_type("None")),
    "__init_subclass__": None,
    "__iter__": ([], path_to_type("typing", "Any")),
    "__le__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__len__": ([], path_to_type("int")),
    "__lt__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__module__": None,
    "__ne__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__new__": None,
    "__next__": ([], path_to_type("typing", "Any")),
    "__reduce__": None,
    "__reduce_ex__": None,
    "__repr__": ([], path_to_type("str")),
    "__setattr__": (
        [path_to_type("str"), path_to_type("typing", "Any")],
        path_to_type("None"),
    ),
    "__setitem__": (
        [path_to_type("typing", "Any"), path_to_type("typing", "Any")],
        path_to_type("typing", "Any"),
    ),
    "__sizeof__": None,
    "__str__": ([], path_to_type("str")),
    "__subclasshook__": None,
}


def module_stubs(module: Any) -> ast.Module:
    types_to_import = {"typing"}
    constants: List[ast.stmt] = []
    classes: List[ast.stmt] = []
    functions: List[ast.stmt] = []
    for member_name, member_value in inspect.getmembers(module):
        element_path = [module.__name__, member_name]
        if member_name in (
            "__builtins__",
            "__cached__",
            "__file__",
            "__loader__",
            "__path__",
            "__spec__",
            "__doc__",
            "__name__",
            "__package__",
        ) or inspect.ismodule(member_value):
            pass
        elif member_name == "__all__":
            constants.insert(
                0,
                ast.Assign(
                    targets=[ast.Name(id=member_name, ctx=ast.Store())],
                    value=ast.Constant(member_value),
                    lineno=0,
                ),
            )
        elif inspect.isclass(member_value):
            classes.append(class_stubs(member_name, member_value, element_path, types_to_import))
        elif inspect.isbuiltin(member_value):
            functions.append(
                function_stub(
                    member_name,
                    member_value,
                    element_path,
                    types_to_import,
                    in_class=False,
                )
            )
        else:
            value_type = type(member_value)
            t = value_type.__qualname__
            if value_type.__module__ != "builtins":
                t = f"{value_type.__module__}.{t}"
            constants.append(
                ast.AnnAssign(
                    target=ast.Name(id=member_name, ctx=ast.Store()),
                    annotation=path_to_type(t),
                    value=ast.Constant(...),
                    simple=1,
                )
            )
    return ast.Module(
        body=[ast.Import(names=[ast.alias(name=t)]) for t in sorted(types_to_import)] + constants + classes + functions,
        type_ignores=[],
    )


def class_stubs(cls_name: str, cls_def: Any, element_path: List[str], types_to_import: Set[str]) -> ast.ClassDef:
    attributes: List[ast.stmt] = []
    methods: List[ast.stmt] = []
    magic_methods: List[ast.stmt] = []
    constants: List[ast.stmt] = []
    for member_name, member_value in inspect.getmembers(cls_def):
        current_element_path = [*element_path, member_name]
        if member_name == "__init__":
            try:
                inspect.signature(cls_def)  # we check it actually exists
                methods = [
                    function_stub(
                        member_name,
                        cls_def,
                        current_element_path,
                        types_to_import,
                        in_class=True,
                    ),
                    *methods,
                ]
            except ValueError as e:
                if "no signature found" not in str(e):
                    raise ValueError(f"Error while parsing signature of {cls_name}.__init_") from e
        elif (member_name in OBJECT_MEMBERS and member_value == OBJECT_MEMBERS[member_name]) or BUILTINS.get(
            member_name, ()
        ) is None:
            pass
        elif inspect.isdatadescriptor(member_value):
            attributes.extend(data_descriptor_stub(member_name, member_value, current_element_path, types_to_import))
        elif inspect.isroutine(member_value):
            (magic_methods if member_name.startswith("__") else methods).append(
                function_stub(
                    member_name,
                    member_value,
                    current_element_path,
                    types_to_import,
                    in_class=True,
                )
            )
        elif member_name == "__match_args__":
            constants.append(
                ast.AnnAssign(
                    target=ast.Name(id=member_name, ctx=ast.Store()),
                    annotation=ast.Subscript(
                        value=path_to_type("tuple"),
                        slice=ast.Tuple(elts=[path_to_type("str"), ast.Constant(...)], ctx=ast.Load()),
                        ctx=ast.Load(),
                    ),
                    value=ast.Constant(member_value),
                    simple=1,
                )
            )
        elif member_value is not None:
            constants.append(
                ast.AnnAssign(
                    target=ast.Name(id=member_name, ctx=ast.Store()),
                    annotation=concatenated_path_to_type(
                        member_value.__class__.__name__, element_path, types_to_import
                    ),
                    value=ast.Constant(...),
                    simple=1,
                )
            )
        else:
            logging.warning(f"Unsupported member {member_name} of class {'.'.join(element_path)}")

    doc = inspect.getdoc(cls_def)
    doc_comment = build_doc_comment(doc) if doc else None
    body = ([doc_comment] if doc_comment else []) + attributes + methods + magic_methods + constants
    if not body:
        body.append(ast.Expr(value=ast.Constant(...)))
    return ast.ClassDef(
        cls_name,
        bases=[],
        keywords=[],
        body=body,
        decorator_list=[path_to_type("typing", "final")],
        type_params=[],
    )


def data_descriptor_stub(
    data_desc_name: str,
    data_desc_def: Any,
    element_path: List[str],
    types_to_import: Set[str],
) -> Tuple[ast.stmt, ...]:
    annotation = None
    doc_comment = None

    doc = inspect.getdoc(data_desc_def)
    if doc is not None:
        annotation = returns_stub(data_desc_name, doc, element_path, types_to_import)
        m = re.findall(r"^ *:return: *(.*) *$", doc, re.MULTILINE)
        if len(m) == 1:
            doc_comment = m[0]
        elif len(m) > 1:
            raise ValueError(
                f"Multiple return annotations found with :return: in {'.'.join(element_path)} documentation"
            )

    assign = ast.AnnAssign(
        target=ast.Name(id=data_desc_name, ctx=ast.Store()),
        annotation=annotation or path_to_type("typing", "Any"),
        simple=1,
    )
    doc_comment = build_doc_comment(doc_comment) if doc_comment else None
    return (assign, doc_comment) if doc_comment else (assign,)


def function_stub(
    fn_name: str,
    fn_def: Any,
    element_path: List[str],
    types_to_import: Set[str],
    *,
    in_class: bool,
) -> ast.FunctionDef:
    body: List[ast.stmt] = []
    doc = inspect.getdoc(fn_def)
    if doc is not None:
        doc_comment = build_doc_comment(doc)
        if doc_comment is not None:
            body.append(doc_comment)

    decorator_list: List[ast.expr] = []
    if in_class and hasattr(fn_def, "__self__"):
        decorator_list.append(ast.Name("staticmethod"))
    if not body:
        body.append(ast.Expr(value=ast.Constant(...)))

    return ast.FunctionDef(
        fn_name,
        arguments_stub(fn_name, fn_def, doc or "", element_path, types_to_import),
        body,
        decorator_list=decorator_list,
        returns=returns_stub(fn_name, doc, element_path, types_to_import) if doc else None,
        type_params=[],
        lineno=0,
    )


def arguments_stub(
    callable_name: str,
    callable_def: Any,
    doc: str,
    element_path: List[str],
    types_to_import: Set[str],
) -> ast.arguments:
    real_parameters: Mapping[str, inspect.Parameter] = inspect.signature(callable_def).parameters
    if callable_name == "__init__":
        real_parameters = {
            "self": inspect.Parameter("self", inspect.Parameter.POSITIONAL_ONLY),
            **real_parameters,
        }

    parsed_param_types = {}
    optional_params = set()

    # Types for magic functions types
    builtin = BUILTINS.get(callable_name)
    if isinstance(builtin, tuple):
        param_names = list(real_parameters.keys())
        if param_names and param_names[0] == "self":
            del param_names[0]
        for name, t in zip(param_names, builtin[0]):
            parsed_param_types[name] = t

    # Types from comment
    for match in re.findall(r"^ *:type *([a-z_]+): ([^\n]*) *$", doc, re.MULTILINE):
        if match[0] not in real_parameters:
            raise ValueError(
                f"The parameter {match[0]} of {'.'.join(element_path)} "
                "is defined in the documentation but not in the function signature"
            )
        type = match[1]
        if type.endswith(", optional"):
            optional_params.add(match[0])
            type = type[:-10]
        parsed_param_types[match[0]] = convert_type_from_doc(type, element_path, types_to_import)

    # we parse the parameters
    posonlyargs = []
    args = []
    vararg = None
    kwonlyargs = []
    kw_defaults: List[Optional[ast.expr]] = []
    kwarg = None
    defaults: List[ast.expr] = []
    for param in real_parameters.values():
        if param.name != "self" and param.name not in parsed_param_types:
            raise ValueError(
                f"The parameter {param.name} of {'.'.join(element_path)} "
                "has no type definition in the function documentation"
            )
        param_ast = ast.arg(arg=param.name, annotation=parsed_param_types.get(param.name))

        default_ast: Optional[ast.expr] = None
        if param.default != param.empty:
            default_ast = ast.Constant(param.default)
            if param.name not in optional_params:
                raise ValueError(
                    f"Parameter {param.name} of {'.'.join(element_path)} "
                    "is optional according to the type but not flagged as such in the doc"
                )
        elif param.name in optional_params:
            raise ValueError(
                f"Parameter {param.name} of {'.'.join(element_path)} "
                "is optional according to the documentation but has no default value"
            )

        if param.kind == param.POSITIONAL_ONLY:
            posonlyargs.append(param_ast)
            if default_ast is not None:
                defaults.append(default_ast)
        elif param.kind == param.POSITIONAL_OR_KEYWORD:
            args.append(param_ast)
            if default_ast is not None:
                defaults.append(default_ast)
        elif param.kind == param.VAR_POSITIONAL:
            vararg = param_ast
        elif param.kind == param.KEYWORD_ONLY:
            kwonlyargs.append(param_ast)
            kw_defaults.append(default_ast)
        elif param.kind == param.VAR_KEYWORD:
            kwarg = param_ast

    return ast.arguments(
        posonlyargs=posonlyargs,
        args=args,
        vararg=vararg,
        kwonlyargs=kwonlyargs,
        kw_defaults=kw_defaults,
        defaults=defaults,
        kwarg=kwarg,
    )


def returns_stub(
    callable_name: str, doc: str, element_path: List[str], types_to_import: Set[str]
) -> Optional[ast.expr]:
    m = re.findall(r"^ *:rtype: *([^\n]*) *$", doc, re.MULTILINE)
    if len(m) == 0:
        builtin = BUILTINS.get(callable_name)
        if isinstance(builtin, tuple) and builtin[1] is not None:
            return builtin[1]
        raise ValueError(
            f"The return type of {'.'.join(element_path)} "
            "has no type definition using :rtype: in the function documentation"
        )
    if len(m) > 1:
        raise ValueError(f"Multiple return type annotations found with :rtype: for {'.'.join(element_path)}")
    return convert_type_from_doc(m[0], element_path, types_to_import)


def convert_type_from_doc(type_str: str, element_path: List[str], types_to_import: Set[str]) -> ast.expr:
    type_str = type_str.strip()
    return parse_type_to_ast(type_str, element_path, types_to_import)


def parse_type_to_ast(type_str: str, element_path: List[str], types_to_import: Set[str]) -> ast.expr:
    # let's tokenize
    tokens = []
    current_token = ""
    for c in type_str:
        if "a" <= c <= "z" or "A" <= c <= "Z" or c == ".":
            current_token += c
        else:
            if current_token:
                tokens.append(current_token)
            current_token = ""
            if c != " ":
                tokens.append(c)
    if current_token:
        tokens.append(current_token)

    # let's first parse nested parenthesis
    stack: List[List[Any]] = [[[]]]
    for token in tokens:
        if token == "[":
            children: List[List[Any]] = [[]]
            stack[-1][-1].append(children)
            stack.append(children)
        elif token == ",":
            stack[-1].append([])
        elif token == "]":
            stack.pop()
        else:
            stack[-1][-1].append(token)

    # then it's easy
    def parse_sequence(sequence: List[Any]) -> ast.expr:
        # we split based on "or"
        or_groups: List[List[str]] = [[]]
        for e in sequence:
            if e == "or":
                or_groups.append([])
            else:
                or_groups[-1].append(e)
        if any(not g for g in or_groups):
            raise ValueError(f"Not able to parse type '{type_str}' used by {'.'.join(element_path)}")

        new_elements: List[ast.expr] = []
        for group in or_groups:
            if len(group) == 1 and isinstance(group[0], str):
                if group[0] == "...":
                    new_elements.append(ast.Constant(...))
                else:
                    new_elements.append(concatenated_path_to_type(group[0], element_path, types_to_import))
            elif len(group) == 2 and isinstance(group[0], str) and isinstance(group[1], list):
                new_elements.append(
                    ast.Subscript(
                        value=concatenated_path_to_type(group[0], element_path, types_to_import),
                        slice=ast.Tuple([parse_sequence(g) for g in group[1]])
                        if len(group[1]) > 1
                        else parse_sequence(group[1][0]),
                        ctx=ast.Load(),
                    )
                )
            elif len(group) == 1 and isinstance(group[0], list):
                if group[0][-1] == ["..."]:
                    # TODO: hack to convert Callable[[T, ...], R] into Callable[..., R]
                    new_elements.append(ast.Constant(...))
                else:
                    new_elements.append(ast.List([parse_sequence(g) for g in group[0] if g]))
            else:
                raise ValueError(f"Not able to parse type '{type_str}' used by {'.'.join(element_path)}")
        return reduce(lambda left, right: ast.BinOp(left=left, op=ast.BitOr(), right=right), new_elements)

    if len(stack[0]) != 1:
        raise ValueError(f"Not able to parse type '{type_str}' used by {'.'.join(element_path)}")
    return parse_sequence(stack[0][0])


def concatenated_path_to_type(path: str, element_path: List[str], types_to_import: Set[str]) -> ast.expr:
    parts = path.split(".")
    if any(not p for p in parts):
        raise ValueError(f"Not able to parse type '{path}' used by {'.'.join(element_path)}")
    if len(parts) > 1:
        types_to_import.add(".".join(parts[:-1]))
    return path_to_type(*parts)


def build_doc_comment(doc: str) -> Optional[ast.Expr]:
    lines = [line.strip() for line in doc.split("\n")]
    clean_lines = []
    for line in lines:
        if line.startswith((":type", ":rtype")):
            continue
        clean_lines.append(line)
    text = "\n".join(clean_lines).strip()
    return ast.Expr(value=ast.Constant(text)) if text else None


def format_with_ruff(file: str) -> None:
    subprocess.check_call([sys.executable, "-m", "ruff", "format", file])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Extract Python type stub from a python module.")
    parser.add_argument("module_name", help="Name of the Python module for which generate stubs")
    parser.add_argument(
        "out",
        help="Name of the Python stub file to write to",
        type=argparse.FileType("wt"),
    )
    parser.add_argument("--ruff", help="Formats the generated stubs using Ruff", action="store_true")
    args = parser.parse_args()
    stub_content = ast.unparse(module_stubs(importlib.import_module(args.module_name))) + "\n" + PROTOCOLS
    args.out.write(stub_content)
    if args.ruff:
        format_with_ruff(args.out.name)



import hashlib
import random
from pathlib import Path

base = Path(__file__).parent.parent
for target, ext in [
    ("sparql_query", "rq"),
    ("sparql_update", "ru"),
    ("sparql_results_xml", "srx"),
    ("sparql_results_json", "srj"),
    ("sparql_results_tsv", "tsv"),
    ("n3", "n3"),
    ("nquads", "nq"),
    ("trig", "trig"),
    ("rdf_xml", "rdf"),
    ("jsonld", "jsonld")
]:
    target_dir = base / "fuzz" / "corpus" / target
    for f in base.rglob(f"*.{ext}"):
        if "manifest" in str(f):
            continue  # we skip the manifests
        with f.open("rb") as fp:
            data = fp.read()
        pos = random.randint(0, len(data))
        data = data[:pos] + b"\xff" + data[pos:]
        hash = hashlib.sha256()
        hash.update(data)
        target_dir.mkdir(parents=True, exist_ok=True)
        with (target_dir / f"{hash.hexdigest()}").open("wb") as fp:
            fp.write(data)



import copy
import math
import pickle
import sys
import unittest

from pyoxigraph import (
    BaseDirection,
    BlankNode,
    DefaultGraph,
    Literal,
    NamedNode,
    Quad,
    Triple,
    Variable,
)

XSD_STRING = NamedNode("http://www.w3.org/2001/XMLSchema#string")
XSD_INTEGER = NamedNode("http://www.w3.org/2001/XMLSchema#integer")
XSD_DOUBLE = NamedNode("http://www.w3.org/2001/XMLSchema#double")
XSD_BOOLEAN = NamedNode("http://www.w3.org/2001/XMLSchema#boolean")
RDF_LANG_STRING = NamedNode("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString")
RDF_DIR_LANG_STRING = NamedNode("http://www.w3.org/1999/02/22-rdf-syntax-ns#dirLangString")


def match_works(test: unittest.TestCase, matched_value: str, constraint: str) -> None:
    """Hack for Python < 3.10 compatibility"""
    if sys.version_info < (3, 10):
        return test.skipTest("match has been introduced by Python 3.10")
    found = True
    exec(
        f"""
match {matched_value}:
    case {constraint}:
        found = True
"""
    )
    test.assertTrue(found)
    return None


class TestNamedNode(unittest.TestCase):
    def test_constructor(self) -> None:
        self.assertEqual(NamedNode("http://foo").value, "http://foo")

    def test_string(self) -> None:
        self.assertEqual(str(NamedNode("http://foo")), "<http://foo>")

    def test_equal(self) -> None:
        self.assertEqual(NamedNode("http://foo"), NamedNode("http://foo"))
        self.assertNotEqual(NamedNode("http://foo"), NamedNode("http://bar"))

    def test_pickle(self) -> None:
        node = NamedNode("http://foo")
        self.assertEqual(pickle.loads(pickle.dumps(node)), node)
        self.assertEqual(copy.copy(node), node)
        self.assertEqual(copy.deepcopy(node), node)

    def test_basic_match(self) -> None:
        match_works(self, 'NamedNode("http://foo")', 'NamedNode("http://foo")')

    def test_wildcard_match(self) -> None:
        match_works(self, 'NamedNode("http://foo")', "NamedNode(x)")


class TestBlankNode(unittest.TestCase):
    def test_constructor(self) -> None:
        self.assertEqual(BlankNode("foo").value, "foo")
        self.assertNotEqual(BlankNode(), BlankNode())

    def test_string(self) -> None:
        self.assertEqual(str(BlankNode("foo")), "_:foo")

    def test_equal(self) -> None:
        self.assertEqual(BlankNode("foo"), BlankNode("foo"))
        self.assertNotEqual(BlankNode("foo"), BlankNode("bar"))
        self.assertNotEqual(BlankNode("foo"), NamedNode("http://foo"))
        self.assertNotEqual(NamedNode("http://foo"), BlankNode("foo"))

    def test_pickle(self) -> None:
        node = BlankNode("foo")
        self.assertEqual(pickle.loads(pickle.dumps(node)), node)
        self.assertEqual(copy.copy(node), node)
        self.assertEqual(copy.deepcopy(node), node)

        auto = BlankNode()
        self.assertEqual(pickle.loads(pickle.dumps(auto)), auto)
        self.assertEqual(copy.copy(auto), auto)
        self.assertEqual(copy.deepcopy(auto), auto)

    def test_basic_match(self) -> None:
        match_works(self, 'BlankNode("foo")', 'BlankNode("foo")')

    def test_wildcard_match(self) -> None:
        match_works(self, 'BlankNode("foo")', "BlankNode(x)")


class TestLiteral(unittest.TestCase):
    def test_constructor(self) -> None:
        self.assertEqual(Literal("foo").value, "foo")
        self.assertEqual(Literal("foo").datatype, XSD_STRING)

        self.assertEqual(Literal("foo", language="en").value, "foo")
        self.assertEqual(Literal("foo", language="en").language, "en")
        self.assertEqual(Literal("foo", language="en").datatype, RDF_LANG_STRING)

        self.assertEqual(Literal("foo", language="en", direction=BaseDirection.LTR).value, "foo")
        self.assertEqual(Literal("foo", language="en", direction=BaseDirection.LTR).language, "en")
        self.assertEqual(Literal("foo", language="en", direction=BaseDirection.LTR).direction, BaseDirection.LTR)
        self.assertEqual(Literal("foo", language="en", direction=BaseDirection.LTR).datatype, RDF_DIR_LANG_STRING)

        self.assertEqual(Literal("foo", datatype=XSD_INTEGER).value, "foo")
        self.assertEqual(Literal("foo", datatype=XSD_INTEGER).datatype, XSD_INTEGER)

        self.assertEqual(Literal(True), Literal("true", datatype=XSD_BOOLEAN))  # noqa: FBT003
        self.assertEqual(Literal(False), Literal("false", datatype=XSD_BOOLEAN))  # noqa: FBT003

        self.assertEqual(Literal(0), Literal("0", datatype=XSD_INTEGER))
        self.assertEqual(Literal(1), Literal("1", datatype=XSD_INTEGER))

        self.assertEqual(Literal(0.1), Literal("0.1", datatype=XSD_DOUBLE))
        self.assertEqual(Literal(math.inf), Literal("INF", datatype=XSD_DOUBLE))
        self.assertEqual(Literal(-math.inf), Literal("-INF", datatype=XSD_DOUBLE))
        self.assertEqual(Literal(-math.nan), Literal("NaN", datatype=XSD_DOUBLE))

    def test_string(self) -> None:
        self.assertEqual(str(Literal("foo")), '"foo"')
        self.assertEqual(str(Literal("foo", language="en")), '"foo"@en')
        self.assertEqual(str(Literal("foo", language="en", direction=BaseDirection.RTL)), '"foo"@en--rtl')
        self.assertEqual(
            str(Literal("foo", datatype=XSD_INTEGER)),
            '"foo"^^<http://www.w3.org/2001/XMLSchema#integer>',
        )

    def test_equals(self) -> None:
        self.assertEqual(Literal("foo", datatype=XSD_STRING), Literal("foo"))
        self.assertEqual(
            Literal("foo", language="en", datatype=RDF_LANG_STRING),
            Literal("foo", language="en"),
        )
        self.assertEqual(
            Literal("foo", language="en", direction=BaseDirection.RTL, datatype=RDF_DIR_LANG_STRING),
            Literal("foo", language="en", direction=BaseDirection.RTL),
        )
        self.assertNotEqual(NamedNode("http://foo"), Literal("foo"))
        self.assertNotEqual(Literal("foo"), NamedNode("http://foo"))
        self.assertNotEqual(BlankNode("foo"), Literal("foo"))
        self.assertNotEqual(Literal("foo"), BlankNode("foo"))

    def test_pickle(self) -> None:
        simple = Literal("foo")
        self.assertEqual(pickle.loads(pickle.dumps(simple)), simple)
        self.assertEqual(copy.copy(simple), simple)
        self.assertEqual(copy.deepcopy(simple), simple)

        lang_tagged = Literal("foo", language="en")
        self.assertEqual(pickle.loads(pickle.dumps(lang_tagged)), lang_tagged)
        self.assertEqual(copy.copy(lang_tagged), lang_tagged)
        self.assertEqual(copy.deepcopy(lang_tagged), lang_tagged)

        dir_lang_tagged = Literal("foo", language="en", direction=BaseDirection.RTL)
        self.assertEqual(pickle.loads(pickle.dumps(dir_lang_tagged)), dir_lang_tagged)
        self.assertEqual(copy.copy(dir_lang_tagged), dir_lang_tagged)
        self.assertEqual(copy.deepcopy(dir_lang_tagged), dir_lang_tagged)

        number = Literal("1", datatype=XSD_INTEGER)
        self.assertEqual(pickle.loads(pickle.dumps(number)), number)
        self.assertEqual(copy.copy(number), number)
        self.assertEqual(copy.deepcopy(number), number)

    def test_basic_match(self) -> None:
        match_works(self, 'Literal("foo", language="en")', 'Literal("foo", language="en")')
        match_works(
            self,
            'Literal("foo", language="en", direction=BaseDirection.RTL)',
            'Literal("foo", language="en", direction=BaseDirection.RTL)',
        )
        match_works(
            self,
            'Literal("1", datatype=XSD_INTEGER)',
            'Literal("1", datatype=NamedNode("http://www.w3.org/2001/XMLSchema#integer"))',
        )

    def test_wildcard_match(self) -> None:
        match_works(self, 'Literal("foo", language="en")', "Literal(v, language=l)")
        match_works(
            self, 'Literal("foo", language="en", direction=BaseDirection.RTL)', "Literal(v, language=l, direction=d)"
        )
        match_works(self, 'Literal("1", datatype=XSD_INTEGER)', "Literal(v, datatype=d)")


class TestBaseDirection(unittest.TestCase):
    def test_constructor(self) -> None:
        self.assertEqual(BaseDirection("ltr"), BaseDirection.LTR)
        self.assertEqual(BaseDirection("rtl"), BaseDirection.RTL)

    def test_string(self) -> None:
        self.assertEqual(str(BaseDirection("ltr")), "ltr")

    def test_equal(self) -> None:
        self.assertEqual(BaseDirection("rtl"), BaseDirection("rtl"))
        self.assertNotEqual(BaseDirection("rtl"), BaseDirection("ltr"))

    def test_pickle(self) -> None:
        v = BaseDirection("rtl")
        self.assertEqual(pickle.loads(pickle.dumps(v)), v)
        self.assertEqual(copy.copy(v), v)
        self.assertEqual(copy.deepcopy(v), v)

    def test_basic_match(self) -> None:
        match_works(self, 'BaseDirection("rtl")', 'BaseDirection("rtl")')

    def test_wildcard_match(self) -> None:
        match_works(self, 'BaseDirection("rtl")', "BaseDirection(x)")


class TestTriple(unittest.TestCase):
    def test_constructor(self) -> None:
        t = Triple(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
        )
        self.assertEqual(t.subject, NamedNode("http://example.com/s"))
        self.assertEqual(t.predicate, NamedNode("http://example.com/p"))
        self.assertEqual(t.object, NamedNode("http://example.com/o"))

    def test_rdf_star_constructor(self) -> None:
        t = Triple(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            Triple(
                NamedNode("http://example.com/os"),
                NamedNode("http://example.com/op"),
                NamedNode("http://example.com/oo"),
            ),
        )
        self.assertEqual(t.subject, NamedNode("http://example.com/s"))
        self.assertEqual(t.predicate, NamedNode("http://example.com/p"))
        self.assertEqual(
            t.object,
            Triple(
                NamedNode("http://example.com/os"),
                NamedNode("http://example.com/op"),
                NamedNode("http://example.com/oo"),
            ),
        )

    def test_mapping(self) -> None:
        t = Triple(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
        )
        self.assertEqual(t[0], NamedNode("http://example.com/s"))
        self.assertEqual(t[1], NamedNode("http://example.com/p"))
        self.assertEqual(t[2], NamedNode("http://example.com/o"))

    def test_destruct(self) -> None:
        (s, p, o) = Triple(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
        )
        self.assertEqual(s, NamedNode("http://example.com/s"))
        self.assertEqual(p, NamedNode("http://example.com/p"))
        self.assertEqual(o, NamedNode("http://example.com/o"))

    def test_string(self) -> None:
        self.assertEqual(
            str(
                Triple(
                    NamedNode("http://example.com/s"),
                    NamedNode("http://example.com/p"),
                    NamedNode("http://example.com/o"),
                )
            ),
            "<http://example.com/s> <http://example.com/p> <http://example.com/o>",
        )

    def test_equals(self) -> None:
        t = Triple(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
        )
        self.assertEqual(
            t,
            Triple(
                NamedNode("http://example.com/s"),
                NamedNode("http://example.com/p"),
                NamedNode("http://example.com/o"),
            ),
        )
        self.assertNotEqual(
            t,
            Triple(
                NamedNode("http://example.com/s"),
                NamedNode("http://example.com/p"),
                NamedNode("http://example.com/s"),
            ),
        )
        self.assertNotEqual(t, NamedNode("http://foo"))
        self.assertNotEqual(t, BlankNode("foo"))
        self.assertNotEqual(t, Literal("foo"))
        self.assertNotEqual(t, DefaultGraph())
        self.assertNotEqual(NamedNode("http://foo"), t)
        self.assertNotEqual(BlankNode("foo"), t)
        self.assertNotEqual(Literal("foo"), t)
        self.assertNotEqual(DefaultGraph(), t)

    def test_pickle(self) -> None:
        triple = Triple(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
        )
        self.assertEqual(pickle.loads(pickle.dumps(triple)), triple)
        self.assertEqual(copy.copy(triple), triple)
        self.assertEqual(copy.deepcopy(triple), triple)

    def test_match(self) -> None:
        match_works(
            self,
            'Triple(NamedNode("http://example.com/s"), NamedNode("http://example.com/p"), '
            'NamedNode("http://example.com/o"))',
            'Triple(NamedNode("http://example.com/s"), NamedNode(p), o)',
        )


class TestDefaultGraph(unittest.TestCase):
    def test_equal(self) -> None:
        self.assertEqual(DefaultGraph(), DefaultGraph())
        self.assertNotEqual(DefaultGraph(), NamedNode("http://bar"))

    def test_pickle(self) -> None:
        self.assertEqual(pickle.loads(pickle.dumps(DefaultGraph())), DefaultGraph())
        self.assertEqual(copy.copy(DefaultGraph()), DefaultGraph())
        self.assertEqual(copy.deepcopy(DefaultGraph()), DefaultGraph())

    def test_match(self) -> None:
        match_works(self, "DefaultGraph()", "DefaultGraph()")


class TestQuad(unittest.TestCase):
    def test_constructor(self) -> None:
        t = Quad(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
            NamedNode("http://example.com/g"),
        )
        self.assertEqual(t.subject, NamedNode("http://example.com/s"))
        self.assertEqual(t.predicate, NamedNode("http://example.com/p"))
        self.assertEqual(t.object, NamedNode("http://example.com/o"))
        self.assertEqual(t.graph_name, NamedNode("http://example.com/g"))
        self.assertEqual(
            t.triple,
            Triple(
                NamedNode("http://example.com/s"),
                NamedNode("http://example.com/p"),
                NamedNode("http://example.com/o"),
            ),
        )
        self.assertEqual(
            Quad(
                NamedNode("http://example.com/s"),
                NamedNode("http://example.com/p"),
                NamedNode("http://example.com/o"),
            ),
            Quad(
                NamedNode("http://example.com/s"),
                NamedNode("http://example.com/p"),
                NamedNode("http://example.com/o"),
                DefaultGraph(),
            ),
        )

    def test_mapping(self) -> None:
        t = Quad(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
            NamedNode("http://example.com/g"),
        )
        self.assertEqual(t[0], NamedNode("http://example.com/s"))
        self.assertEqual(t[1], NamedNode("http://example.com/p"))
        self.assertEqual(t[2], NamedNode("http://example.com/o"))
        self.assertEqual(t[3], NamedNode("http://example.com/g"))

    def test_destruct(self) -> None:
        (s, p, o, g) = Quad(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
            NamedNode("http://example.com/g"),
        )
        self.assertEqual(s, NamedNode("http://example.com/s"))
        self.assertEqual(p, NamedNode("http://example.com/p"))
        self.assertEqual(o, NamedNode("http://example.com/o"))
        self.assertEqual(g, NamedNode("http://example.com/g"))

    def test_string(self) -> None:
        self.assertEqual(
            str(
                Triple(
                    NamedNode("http://example.com/s"),
                    NamedNode("http://example.com/p"),
                    NamedNode("http://example.com/o"),
                )
            ),
            "<http://example.com/s> <http://example.com/p> <http://example.com/o>",
        )

    def test_pickle(self) -> None:
        quad = Quad(
            NamedNode("http://example.com/s"),
            NamedNode("http://example.com/p"),
            NamedNode("http://example.com/o"),
            NamedNode("http://example.com/g"),
        )
        self.assertEqual(pickle.loads(pickle.dumps(quad)), quad)
        self.assertEqual(copy.copy(quad), quad)
        self.assertEqual(copy.deepcopy(quad), quad)

    def test_match(self) -> None:
        match_works(
            self,
            'Quad(NamedNode("http://example.com/s"), NamedNode("http://example.com/p"), '
            'NamedNode("http://example.com/o"), NamedNode("http://example.com/g"))',
            'Quad(NamedNode("http://example.com/s"), NamedNode(p), o, NamedNode("http://example.com/g"))',
        )


class TestVariable(unittest.TestCase):
    def test_constructor(self) -> None:
        self.assertEqual(Variable("foo").value, "foo")

    def test_string(self) -> None:
        self.assertEqual(str(Variable("foo")), "?foo")

    def test_equal(self) -> None:
        self.assertEqual(Variable("foo"), Variable("foo"))
        self.assertNotEqual(Variable("foo"), Variable("bar"))

    def test_pickle(self) -> None:
        v = Variable("foo")
        self.assertEqual(pickle.loads(pickle.dumps(v)), v)
        self.assertEqual(copy.copy(v), v)
        self.assertEqual(copy.deepcopy(v), v)

    def test_basic_match(self) -> None:
        match_works(self, 'Variable("foo")', 'Variable("foo")')

    def test_wildcard_match(self) -> None:
        match_works(self, 'Variable("foo")', "Variable(x)")


if __name__ == "__main__":
    unittest.main()



import re
import sys
from pathlib import Path
from urllib.error import HTTPError
from urllib.parse import urlsplit, urlunsplit
from urllib.request import urlopen

LINK_REGEXES = {
    r"\[[^]]+]\((https?://(w3c.github.io|www.w3.org)/[^)]+)\)",  # Markdown
    r"<(https?://(w3c.github.io|www.w3.org)/[^>]+)>`_",  # reStructuredText
}

base_path = Path(__file__).parent.parent
spec_cache = {}
errors = set()

for ext in ("md", "rs", "rst"):
    for file in base_path.rglob(f"*.{ext}"):
        content = file.read_text()
        for link_regex in LINK_REGEXES:
            for m in re.finditer(link_regex, content):
                url = m.group(1)
                (scheme, host, path, query, fragment) = urlsplit(url)
                if scheme != "https":
                    errors.add(f"HTTP URL used by {url} in {file}")
                if query != "":
                    errors.add(f"URL query used by {url} in {file}")
                if path.endswith(".html/"):
                    errors.add(f".html/ used by {url} in {file}")
                base_url = urlunsplit(("https", host, path.rstrip("/"), "", ""))
                if base_url not in spec_cache:
                    try:
                        with urlopen(base_url) as response:
                            spec_cache[base_url] = response.read().decode(
                                errors="ignore"
                            )
                    except HTTPError as e:
                        errors.add(
                            f"Fetching {url} used in {file} return HTTP error: {e}"
                        )
                spec_content = spec_cache.get(base_url, "")
                if (
                    fragment != ""
                    and re.search(rf"[iI][dD]\s*=\s*['\"]{fragment}['\"]", spec_content)
                    is None
                ):
                    errors.add(
                        f"Fragment {fragment} of {url} used in {file} does not exist"
                    )

print("Used specs:")
for url in sorted(spec_cache.keys()):
    print(url)

if errors:
    print()
    for error in sorted(errors):
        print(error, file=sys.stderr)
    exit(1)
